package com.capitalware.ufm.action.objects;

import java.util.ArrayList;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.handlers.Tokens;
import com.capitalware.ufm.mq.objects.QueueObject;
import com.capitalware.ufm.objects.CWConstants;
import com.capitalware.ufm.objects.Cryptography;

/**
 * This class holds the values for the Receive action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class MQReceiveObject extends UFMFileMQObject
{
   public QueueObject         backOutQName = null;
   public String              defaultDirectory = CWConstants.HOME_PATH;
   public String              defaultFilename = null;
   public String              defaultJobXmlFile = null;
   public char                processType = CWConstants.ONE_PASS;
   public boolean             getWithConvert = false;
   public boolean             overrideDirectoryName = false;
   public boolean             overrideFileName = false;
   public boolean             overrideJobName = false;
   public boolean             createDir = false;
   public boolean             append = false;
   public Cryptography        crypto = null;
   public ArrayList<UFMObject> actionAL = null;

   /**
    * The constructor
    */
   public MQReceiveObject()
   {
      super();
   }

   /**
    * Process tokens in Action Object's variables
    */
   public void performTokenization()
   {
      if (backOutQName != null)
      {
         if (backOutQName.qName != null)
            backOutQName.qName = Tokens.replace(backOutQName.qName, paClass.actionProperties);
         
         if (backOutQName.qMgrName != null)
            backOutQName.qMgrName = Tokens.replace(backOutQName.qMgrName, paClass.actionProperties);
      }
   }

   /**
    * Check the user specified parameters.
    */
   public void checkObject()
   {
      if ( overrideJobName && (defaultJobXmlFile == null))
         throw new IllegalArgumentException("Error: You must specify a Default / XmlJobFile to use the 'override' attribute.");

      if ( overrideFileName && (defaultFilename == null))
         throw new IllegalArgumentException("Error: You must specify a Default / FileName to use the 'override' attribute.");

      if (crypto != null)
      {
         if ( (crypto.keySize != 128) && (crypto.keySize != 192) && (crypto.keySize != 256) )
            throw new IllegalArgumentException("Error: You must specify a keysize attribute that is 128 or 192 or 256 for the Send action.");

         if ("".equals(crypto.passPhrase))
            throw new IllegalArgumentException("Error: PassPhrase attribute cannot be blank for the Send action.");
      }
   }

   /**
    * Print out the user supplied parameters.
    */
   public void printValues()
   {
      UFM.logger.info("     GetWithConvert       = " + getWithConvert);
      UFM.logger.info("     BackOutQName         = " + (backOutQName==null?"":backOutQName));
      UFM.logger.info("     Default JobFile      = " + (defaultJobXmlFile==null?"":defaultJobXmlFile));
      UFM.logger.info("     Default Directory Path= " + defaultDirectory);
      UFM.logger.info("     Default Filename     = " + (defaultFilename==null?"":defaultFilename));
      UFM.logger.info("     Override JobXmlFile  = " + overrideJobName);
      UFM.logger.info("     Override Directory   = " + overrideDirectoryName);
      UFM.logger.info("     Override File Name   = " + overrideFileName);
      UFM.logger.info("     Append               = " + append);
      UFM.logger.info("     Create Directory     = " + createDir);

      String type = "Invalid";
      if (processType == CWConstants.ONE_PASS)
         type = "OnePass";
      else if (processType == CWConstants.RUN_UNTIL_EMPTY)
         type = "RunUntilEmpty";
      else if (processType == CWConstants.TRIGGERED)
         type = "Triggered";
      else if (processType == CWConstants.DAEMON)
         type = "Daemon";

      UFM.logger.info("     Processing Type      = " + type);
      
      if (crypto != null)
      {
         UFM.logger.info("     KeySize            = " + crypto.keySize);
         UFM.logger.info("     PassPhrase         = " + crypto.passPhrase );
      }
      
      if (actionAL != null)
      {
         UFM.logger.info("     MQReceive Actions to be processed:");
         for (int i=0; i < actionAL.size(); i++)
         {
            UFMObject fmo = (UFMObject)actionAL.get(i);
            UFM.logger.info("       Action ["+(i+1)+"]         = " + fmo.actionName);
         }
      }
}
   
   /**
    * Convert the object to a string.
    * @return s
    */
   public ArrayList<String> toStringArray()
   {
      ArrayList<String> al = new ArrayList<String>();

      al.add("<"+CWConstants.MQ_RECEIVE_ELEMENT+" "+CWConstants.GETWITHCONVERT_ATTRIBUTE+"=\""+getWithConvert+"\" "+CWConstants.RUN_ATTRIBUTE+"=\""+processType+"\" "+CWConstants.APPEND_ATTRIBUTE+"=\""+append+"\""+CWConstants.CREATE_DIR_ATTRIBUTE+"=\""+createDir+"\">"+LINE_SEPARATOR);

      /*
       * MQ info
       */
      ArrayList<String> mq = createMQTag();
      for (int i=0; i < mq.size(); i++)
         al.add(mq.get(i));
      
      /*
       * Default info
       */
      if ( (defaultJobXmlFile != null) || (defaultDirectory != null) || (defaultFilename != null) )
      {
         al.add("  <"+CWConstants.DEFAULT_ELEMENT+">"+LINE_SEPARATOR);

         if (defaultJobXmlFile != null)
            al.add("    <"+CWConstants.EXECUTE_ELEMENT+">"+defaultJobXmlFile+"</"+CWConstants.EXECUTE_ELEMENT+">"+LINE_SEPARATOR);

         if (defaultDirectory != null)
            al.add("    <"+CWConstants.DIRECTORY_ELEMENT+">"+defaultDirectory+"</"+CWConstants.DIRECTORY_ELEMENT+">"+LINE_SEPARATOR);

         if (defaultFilename != null)
            al.add("    <"+CWConstants.FILENAME_ELEMENT+">"+defaultFilename+"</"+CWConstants.FILENAME_ELEMENT+">"+LINE_SEPARATOR);
         
         al.add("  </"+CWConstants.DEFAULT_ELEMENT+">"+LINE_SEPARATOR);
      }
      
      /*
       * Add info from actions
       */
      if (actionAL != null)
      {
         for (int i=0; i < actionAL.size(); i++)
         {
            UFMObject fmo = (UFMObject)actionAL.get(i);
            ArrayList<String> actionAL = fmo.toStringArray();
            for (int j=0; j < actionAL.size(); j++)
            {
               al.add("  "+actionAL.get(j));
            }
         }
      }

      al.add("</"+CWConstants.MQ_RECEIVE_ELEMENT+">"+LINE_SEPARATOR);
      
      return al;
   }

   /**
    * Retrieve the values in this action for the status update.
    * @return s
    */
   public String toStringForStatusUpdate(String event)
   {
      String s;

      s = CWConstants.MQ_RECEIVE_ELEMENT+", "+labelName+", "+event+", "+createMQTagForStatusUpdate()+CWConstants.CREATE_DIR_ATTRIBUTE+"="+createDir+", "+CWConstants.GETWITHCONVERT_ATTRIBUTE+"="+getWithConvert+", "+CWConstants.RUN_ATTRIBUTE+"="+processType+", "+CWConstants.APPEND_ATTRIBUTE+"="+append+LINE_SEPARATOR;

      return s;
   }
}
