package com.capitalware.ufm.actions;

import java.io.File;
import java.io.IOException;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.objects.ConvertObject;
import com.capitalware.ufm.handlers.CharacterConversion;
import com.capitalware.ufm.handlers.IO;
import com.capitalware.ufm.handlers.Tokens;
import com.capitalware.ufm.objects.CWConstants;
import com.capitalware.ufm.objects.UFMException;

/**
 * This class will handle the convert action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class ConvertAction extends UFMAction
{
   private ConvertObject   co;

   /**
    * The constructor
    * @param co
    * @param logger
    */
   public ConvertAction(ConvertObject co)
   {
      super(co.actionName);
      this.co = co;
   }

   /**
    * Perform the convert work.
    * @return exitCode
    */
   public int doWork()
   {
      int exitCode = 0;
      co.printValues();

      if (co.srcFilename.indexOf('*') != -1)
         exitCode = handleManyFile();
      else
         exitCode = handleOneFile();

      if (!co.onErrorFail)
         exitCode = 0; /* reset back to zero */
      
      return exitCode;
   }

   /**
    * Convert a single file
    * @return exitCode
    */
   private int handleOneFile()
   {
      int exitCode = 0;
      String toPathAndFilename;
      String toFilename;

      if (co.toFile != null)
         toFilename = IO.getFileName(co.toFile);
      else         
         toFilename = IO.getFileName(co.srcFilename);

      toPathAndFilename = IO.apendFileToPath(co.toDir, toFilename);

      toPathAndFilename = Tokens.replace(co.actionName, toPathAndFilename, IO.getFileName(co.srcFilename), 1, co.paClass.actionProperties);

      exitCode = handleConversion(co.srcFilename, toPathAndFilename);

      return exitCode;
   }

   /**
    * Convert many files
    * @return exitCode
    */
   private int handleManyFile()
   {
      int exitCode = 0;
      String toPathAndFilename;
      File[] fileList;
      
      try
      {
         fileList = IO.getFilteredFileList(co.srcFilename);
         
         for (int i=0; i < fileList.length; i++)
         {
            String toFilename;

            if (co.toFile == null)
               toFilename = fileList[i].getName();
            else
               toFilename = co.toFile;

            toPathAndFilename = IO.apendFileToPath(co.toDir, toFilename);

            toPathAndFilename = Tokens.replace(co.actionName, toPathAndFilename, fileList[i].getName(), i+1, co.paClass.actionProperties);

            exitCode = handleConversion(fileList[i].toString(), toPathAndFilename);
         }
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 8;
      }

      return exitCode;
   }

   /**
    * Select the appropriate character conversion.
    * @return exitCode
    */
   private int handleConversion(String src, String tgt)
   {
      int exitCode = 0;

      try
      {
         if (co.mode == CWConstants.CONVERT_DOS_2_EBCDIC)
            CharacterConversion.dos2Ebcdic(src, tgt);
         else if (co.mode == CWConstants.CONVERT_DOS_2_MAC)
            CharacterConversion.dos2Mac(src, tgt);
         else if (co.mode == CWConstants.CONVERT_DOS_2_UNIX)
            CharacterConversion.dos2Unix(src, tgt);
         else if (co.mode == CWConstants.CONVERT_EBCDIC_2_DOS)
            CharacterConversion.ebcdic2Dos(src, tgt);
         else if (co.mode == CWConstants.CONVERT_EBCDIC_2_MAC)
            CharacterConversion.ebcdic2Mac(src, tgt);
         else if (co.mode == CWConstants.CONVERT_EBCDIC_2_UNIX)
            CharacterConversion.ebcdic2Unix(src, tgt);
         else if (co.mode == CWConstants.CONVERT_MAC_2_DOS)
            CharacterConversion.mac2Dos(src, tgt);
         else if (co.mode == CWConstants.CONVERT_MAC_2_EBCDIC)
            CharacterConversion.mac2Ebcdic(src, tgt);
         else if (co.mode == CWConstants.CONVERT_MAC_2_UNIX)
            CharacterConversion.mac2Unix(src, tgt);
         else if (co.mode == CWConstants.CONVERT_UNIX_2_DOS)
            CharacterConversion.unix2Dos(src, tgt);
         else if (co.mode == CWConstants.CONVERT_UNIX_2_EBCDIC)
            CharacterConversion.unix2Ebcdic(src, tgt);
         else if (co.mode == CWConstants.CONVERT_UNIX_2_MAC)
            CharacterConversion.unix2Mac(src, tgt);

        UFM.logger.info("Successfully converted ["+CWConstants.CONVERT_ARRAY[co.mode]+"] file '"+src+ "' to '" + tgt + "'");
         
         co.fileCount++;
      }
      catch (UFMException e)
      {
         exitCode = 8;
      }

      return exitCode;
   }
}
