package com.capitalware.ufm.actions;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;

import org.jdom.JDOMException;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.objects.FtpObject;
import com.capitalware.ufm.ftp.FTPConnection;
import com.capitalware.ufm.ftp.objects.*;
import com.capitalware.ufm.handlers.IO;
import com.capitalware.ufm.handlers.InternalMessageQueue;
import com.capitalware.ufm.handlers.Tokens;
import com.capitalware.ufm.internal.messages.UpdateStatus;
import com.capitalware.ufm.objects.ConnectionObject;
import com.capitalware.ufm.objects.CWConstants;
import com.capitalware.ufm.xml.FtpXml;

/**
 * This class will handle the FTP action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class FtpAction extends UFMAction
{
   private     FtpObject      fo = null;
   private     FtpXml         fxh = null;
   private     InternalMessageQueue   iq = null;
   private     String         fileSeparator = "/"; 

   /**
    * The constructor
    * @param fo
    */
   public FtpAction(FtpObject fo, InternalMessageQueue iq)
   {
      super(fo.actionName);
      this.fo = fo;
      this.iq = iq;
   }

   /**
    * Perform the work.
    * @return exitCode
    */
   public int doWork()
   {
      int                     exitCode = 0;
      ConnectionObject        co = null;
      ArrayList<FtpCmdObject> cmdsAL = null;
      FTPConnection           fc = null;
      
      fo.printValues();
      
      fxh = new FtpXml();
      try
      {
         fxh.readXMLFile(fo.xmlFile);
         
         co = fxh.getFtpConnection();
         cmdsAL = fxh.getFtpCmds();

         /* Dump out the values */
         co.printValues("FTP");
         printValues(cmdsAL);

         /* Check/validate the values */
         co.checkObject("FTP");
         checkObject(cmdsAL);

         /* connect to the FTP server. */
         fc = new FTPConnection();
         fc.connect(co.hostname, co.port);
         UFM.logger.info("Connected to " + co.hostname + " on port # " + co.port);

         /* log into the FTP server */
         fc.login(co.userID, co.password);
         UFM.logger.info("Logged in with " + co.userID);
         
         String curDir = fc.getCurrentDirectory();
         UFM.logger.debug("Current directory is '" + curDir+"'");
         
         if (curDir.indexOf('\\') != -1) /* windows FTP server? */
            fileSeparator = "\\";
         
         /* Process FTP commands. */
         exitCode = processFtpCommands(fc, cmdsAL);
      }
      catch (JDOMException e)
      {
         UFM.logger.fatal("JDOMException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 8;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 8;
      }
      finally
      {
         if (fc != null)
         {
            fc.disconnect();
            UFM.logger.info("Disconnected from " + co.hostname);
         }
         
         if (!fo.onErrorFail)
            exitCode = 0; /* reset back to zero */
      }
      
      return exitCode;
   }
   
   
   /**
    * Print out FTP command objects, if any.
    * @param cmdsAL
    */
   private void printValues(ArrayList<FtpCmdObject> cmdsAL)
   {
      if (cmdsAL != null)
      {
         UFM.logger.info("     FTP commands to be processed:");
         for (int i=0; i < cmdsAL.size(); i++)
         {
            UFM.logger.info("       FTP Command ["+(i+1)+"]: " + ((FtpCmdObject)cmdsAL.get(i)).cmdName);
            ((FtpCmdObject)cmdsAL.get(i)).printValues();
         }
      }
      
   }
   
   /**
    * Issue checkObject for each FTP command object
    * @param cmdsAL
    */
   private void checkObject(ArrayList<FtpCmdObject> cmdsAL)
   {
      if (cmdsAL != null)
      {
         for (int i=0; i < cmdsAL.size(); i++)
         {
            ((FtpCmdObject)cmdsAL.get(i)).performTokenization(fo.paClass.actionProperties);
            ((FtpCmdObject)cmdsAL.get(i)).checkObject();
         }
      }
   }
   
   /**
    * Loop through FTP commands and process them.
    * @param fc
    * @param cmdsAL
    * @return exitCode
    * @throws IOException
    */
   private int processFtpCommands(FTPConnection fc, ArrayList<FtpCmdObject> cmdsAL) throws IOException
   {
      SimpleDateFormat dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      int              exitCode = 0;
      FtpCmdObject     ftpco = null;
      
      /**
       * Process FTP commands.
       */
      boolean cmdResults = true;
      if (cmdsAL != null)
      {
         for (int i=0; i < cmdsAL.size(); i++)
         {
            ftpco = cmdsAL.get(i);
            if (ftpco instanceof GetObject)
            {
               exitCode = processGetCommand(fc, (GetObject)ftpco);
            }
            else if (ftpco instanceof PutObject)
            {
               exitCode = processPutCommand( fc, (PutObject)ftpco);
            }
            else if (ftpco instanceof CdObject)
            {
               CdObject co = (CdObject)ftpco;
               cmdResults = fc.changeDirectory(co.targetDir);
               if (cmdResults)
                  UFM.logger.info("Changed directory to " + co.targetDir);

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fo.actionName+", "+fo.labelName+", FTP_Command=cd "+co.targetDir+"\n"));
            }
            else if (ftpco instanceof TypeObject)
            {
               TypeObject to = (TypeObject)ftpco;
               if (to.mode == CWConstants.FTP_MODE_ASCII)
                  cmdResults = fc.ascii();
               else if (to.mode == CWConstants.FTP_MODE_BINARY)
                  cmdResults = fc.binary();
               
               if (cmdResults)
                  UFM.logger.info("Changed typed to " + (to.mode == CWConstants.FTP_MODE_BINARY?"Binary":"ASCII"));

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fo.actionName+", "+fo.labelName+", FTP_Command="+(to.mode == CWConstants.FTP_MODE_BINARY?"Binary":"ASCII")+"\n"));
            }
            else if (ftpco instanceof MkDirObject)
            {
               MkDirObject mdo = (MkDirObject)ftpco;
               cmdResults = fc.makeDirectory(mdo.remoteDir);
               if (cmdResults)
                  UFM.logger.info("Created directory: " + mdo.remoteDir);

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fo.actionName+", "+fo.labelName+", FTP_Command=mkdir "+mdo.remoteDir+"\n"));
            }
            else if (ftpco instanceof RmDirObject)
            {
               RmDirObject rdo = (RmDirObject)ftpco;
               cmdResults = fc.removeDirectory(rdo.remoteDir);
               if (cmdResults)
                  UFM.logger.info("Deleted directory: " + rdo.remoteDir);

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fo.actionName+", "+fo.labelName+", FTP_Command=rmdir "+rdo.remoteDir+"\n"));
            }
            else if (ftpco instanceof DeleteObject)
            {
               DeleteObject dlo = (DeleteObject)ftpco;
               String deletedFile = IO.apendFileToPath(dlo.remoteDir, dlo.remoteFile, "/");
               cmdResults = fc.deleteFile(deletedFile);
               if (cmdResults)
                  UFM.logger.info("Deleted file: " + deletedFile);

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fo.actionName+", "+fo.labelName+", FTP_Command=del "+deletedFile+"\n"));
            }
            
            if (!cmdResults)
            {
               exitCode = 8;
               break;
            }
         }
      }
      
      return exitCode;
   }
   
   /**
    * Process the get command
    * @param fc
    * @param go
    * @return exitCode
    * @throws IOException
    */
   private int processGetCommand(FTPConnection fc, GetObject go) throws IOException
   {
      SimpleDateFormat dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      int      exitCode = 0;
      boolean  cmdResults = true;
      String   remoteFile; 
      String   localFile;

      if (go.remoteFilename.indexOf('*') != -1)
      {
         FtpObjectList fol = fc.getListOfItems(go.remoteFilename);
         
         for (int i=0;i < fol.files.size(); i++)
         {
            remoteFile = IO.apendFileToPath(go.remoteDir, fol.files.get(i), fileSeparator);
            localFile = IO.apendFileToPath(go.localDir, (go.localFilename == null?fol.files.get(i):go.localFilename));
            cmdResults = fc.downloadFile(remoteFile, localFile);
            if (cmdResults)
            {
               UFM.logger.info("Retrieved file from " + remoteFile + " and stored as " + localFile);
               go.fileCount++;

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fo.actionName+", "+fo.labelName+", FTP_Command=get "+remoteFile+"\n"));
            }
         }
      }
      else
      {
         remoteFile = IO.apendFileToPath(go.remoteDir, go.remoteFilename, fileSeparator);
         localFile = IO.apendFileToPath(go.localDir, (go.localFilename == null?go.remoteFilename:go.localFilename));
         cmdResults = fc.downloadFile(remoteFile, localFile);
         if (cmdResults)
         {
            UFM.logger.info("Retrieved file from " + remoteFile + " and stored as " + localFile);
            go.fileCount++;

            if (iq != null)
               iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fo.actionName+", "+fo.labelName+", FTP_Command=get "+remoteFile+"\n"));
         }
      }
      
      return exitCode;
   }
   
   /**
    * Process the put command
    * @param fc
    * @param po
    * @return exitCode
    * @throws IOException
    */
   private int processPutCommand(FTPConnection fc, PutObject po) throws IOException
   {
      SimpleDateFormat dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      int      exitCode = 0;
      boolean  cmdResults = true;
      String   remoteFile; 
      String   localFile;

      if (po.localFilename.indexOf('*') != -1)
      {
         File[] fileList;
         
         try
         {
            if (po.localDir != null)
               fileList = IO.getFilteredFileList(IO.apendFileToPath(po.localDir, po.localFilename));
            else
               fileList = IO.getFilteredFileList(po.localFilename);
            
            for (int i=0; i < fileList.length; i++)
            {
               String toFilename = fileList[i].getName();

               remoteFile = IO.apendFileToPath(po.remoteDir, toFilename, fileSeparator);
               remoteFile = Tokens.replace("put", remoteFile, toFilename, i+1, fo.paClass.actionProperties);
               
               localFile = IO.apendFileToPath(po.localDir, toFilename);
               cmdResults = fc.uploadFile(remoteFile, localFile);
               if (cmdResults)
               {
                  UFM.logger.info("Stored file " + localFile + " to " + remoteFile);
                  po.fileCount++;

                  if (iq != null)
                     iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fo.actionName+", "+fo.labelName+", FTP_Command=put "+remoteFile+"\n"));
               }
            }
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            UFM.globalErrorText = e.getLocalizedMessage();
            exitCode = 8;
         }
      }
      else
      {
         remoteFile = IO.apendFileToPath(po.remoteDir, (po.remoteFilename == null?po.localFilename:po.remoteFilename), fileSeparator);
         localFile = IO.apendFileToPath(po.localDir, po.localFilename);
         cmdResults = fc.uploadFile(remoteFile, localFile);
         if (cmdResults)
         {
            UFM.logger.info("Stored file " + localFile + " to " + remoteFile);
            po.fileCount++;

            if (iq != null)
               iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fo.actionName+", "+fo.labelName+", FTP_Command=put "+remoteFile+"\n"));
         }
      }
      
      return exitCode;
   }
}
