package com.capitalware.ufm.actions;

import java.io.File;
import java.io.IOException;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.objects.MoveObject;
import com.capitalware.ufm.handlers.IO;
import com.capitalware.ufm.handlers.Tokens;
import com.capitalware.ufm.objects.UFMException;

/**
 * This class will handle the move action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class MoveAction extends UFMAction
{
   private MoveObject   mo;

   /**
    * The constructor
    * @param mo
    */
   public MoveAction(MoveObject mo)
   {
      super(mo.actionName);
      this.mo = mo;
   }

   /**
    * Perform the work.
    * @return exitCode
    */
   public int doWork()
   {
      int exitCode = 0;
      mo.printValues();

      if (mo.srcFilename.indexOf('*') != -1)
         exitCode = handleManyFile();
      else
         exitCode = handleOneFile();

      mo.paClass.handleDeletedWatchFoundFile();

      if (!mo.onErrorFail)
         exitCode = 0; /* reset back to zero */

      return exitCode;
   }

   /**
    * Move a single file
    * @return exitCode
    */
   private int handleOneFile()
   {
      int exitCode = 0;
      String toPathAndFilename;
      String toFilename;

      if (mo.toFile != null)
         toFilename = IO.getFileName(mo.toFile);
      else
         toFilename = IO.getFileName(mo.srcFilename);

      toPathAndFilename = IO.apendFileToPath(mo.toDir, toFilename);

      toPathAndFilename = Tokens.replace(mo.actionName, toPathAndFilename, IO.getFileName(mo.srcFilename), 1, mo.paClass.actionProperties);

      try
      {
         IO.copyFile(mo.srcFilename, toPathAndFilename);
         
         try
         {
            IO.deleteFile(mo.srcFilename);
         }
         catch (UFMException e)
         {
            if (mo.onErrorFail)
               throw e;
            else
               UFM.handleOnError(mo.paClass.actionProperties, null);
         }
         
         mo.fileCount++;
      }
      catch (UFMException e)
      {
         exitCode = 8;
      }

      return exitCode;
   }

   /**
    * Move a many files
    * @return exitCode
    */
   private int handleManyFile()
   {
      int exitCode = 0;
      String toPathAndFilename;
      File[] fileList;
      
      try
      {
         fileList = IO.getFilteredFileList(mo.srcFilename);

         for (int i=0; i < fileList.length; i++)
         {
            String toFilename;

            if (mo.toFile == null)
               toFilename = fileList[i].getName();
            else
               toFilename = mo.toFile;

            toPathAndFilename = IO.apendFileToPath(mo.toDir, toFilename);

            toPathAndFilename = Tokens.replace(mo.actionName, toPathAndFilename, fileList[i].getName(), i+1, mo.paClass.actionProperties);

            IO.copyFile(fileList[i].toString(), toPathAndFilename);
            
            try
            {
               IO.deleteFile(fileList[i].toString());
            }
            catch (UFMException e)
            {
               if (mo.onErrorFail)
                  throw e;
               else
                  UFM.handleOnError(mo.paClass.actionProperties, null);
            }

            mo.fileCount++;
         }
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 8;
      }
      catch (UFMException e)
      {
         exitCode = 8;
      }

      return exitCode;
   }
}
