package com.capitalware.ufm.actions;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.objects.ReplaceTextObject;
import com.capitalware.ufm.handlers.IO;
import com.capitalware.ufm.handlers.Tokens;

/**
 * This class will handle the ReplaceText action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class ReplaceTextAction extends UFMAction
{
   private final static String   LineSeparator = System.getProperty("line.separator");
   private ReplaceTextObject     rto;

   /**
    * The constructor
    * @param rto
    */
   public ReplaceTextAction(ReplaceTextObject rto)
   {
      super(rto.actionName);
      this.rto = rto;
   }

   /**
    * Perform the work.
    * @return exitCode
    */
   public int doWork()
   {
      int exitCode = 0;
      String toPathAndFilename;
      RandomAccessFile in = null;
      RandomAccessFile out = null;
      FileChannel srcChannel = null;
      FileChannel dstChannel = null;
      FileLock srcLock = null;
      FileLock dstLock = null;
      long   len;
      rto.printValues();

      toPathAndFilename = IO.apendFileToPath(rto.toDir, rto.toFile);
      toPathAndFilename = Tokens.replace(rto.actionName, toPathAndFilename, IO.getFileName(rto.srcFilename), 1, rto.paClass.actionProperties);

      try
      {
         in = new RandomAccessFile(rto.srcFilename, "r");
         srcChannel = in.getChannel();
         len = in.length();
         srcLock = srcChannel.tryLock(0, len, true); /* shared lock for reading */
         
         if (srcLock != null)
         {
            out = new RandomAccessFile(toPathAndFilename, "rw");
            dstChannel = out.getChannel();
            dstLock = dstChannel.tryLock(); /* exclusive lock for writing */
            
            if (dstLock != null)
            {
               loopThroughFile(in,  out);
               rto.fileCount++;
               
               UFM.logger.info("Successfully replaced text in '" + rto.srcFilename + "' to file '" + toPathAndFilename + "'");
            }
            else
            {
               UFM.logger.fatal("ReplaceTextAction cannot acquire lock for file: " + toPathAndFilename);
               UFM.globalErrorText = "ReplaceTextAction cannot acquire lock for file: " + toPathAndFilename;
               exitCode = 8;
            }
         }
         else
         {
            UFM.logger.fatal("ReplaceTextAction cannot acquire lock for file: " + rto.srcFilename);
            UFM.globalErrorText = "ReplaceTextAction cannot acquire lock for file: " + rto.srcFilename;
            exitCode = 8;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 8;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 8;
      }
      finally
      {
         try
         {
            if(dstLock != null)
               dstLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            exitCode = 8;
         }

         try
         {
            if(srcLock != null)
               srcLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            exitCode = 8;
         }
         // Close the channels
         try
         {
            if (srcChannel != null)
               srcChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            exitCode = 8;
         }
         
         try
         {
            if (dstChannel != null)
               dstChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            exitCode = 8;
         }
         try
         {
            if (in != null)
               in.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            exitCode = 8;
         }
         try
         {
            if (out != null)
               out.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            exitCode = 8;
         }
      }

      if (!rto.onErrorFail)
         exitCode = 0; /* reset back to zero */

      return exitCode;
   }

   /**
    * Loop through the file and perform the search and replace.
    * @param in
    * @param out
    * @throws IOException
    */
   private void loopThroughFile(RandomAccessFile in, RandomAccessFile out) throws IOException
   {
      String line;
      String prefix;
      String data;
      String postfix;
      long   len;
      int i = 0;
      
      while ((line = in.readLine()) != null)
      {
         if (i >= (rto.row.start - 1))
         {
            if ( (rto.row.end == -1) ||  
                 ( (rto.row.end != -1) && (i < rto.row.end) ) )
            {
               if ( (rto.column.start == 1) && (rto.column.end == -1) )
                  line = line.replaceAll(rto.findText, rto.replaceText);
               else
               {
                  len = line.length();
                  
                  if (rto.column.start > len)
                  {
                     // do nothing - starting position beyond end of line.
                  }
                  else
                  {
                     prefix = line.substring(0, rto.column.start-1);
                     
                     if ( (rto.column.end < len) && (rto.column.end != -1) )
                        postfix = line.substring(rto.column.end);
                     else
                        postfix = "";
                     
                     if ( (rto.column.end == -1) || (rto.column.end > len) )
                        data = line.substring(rto.column.start-1);
                     else
                        data = line.substring(rto.column.start-1, rto.column.end);
                     
                     line = prefix + data.replaceAll(rto.findText, rto.replaceText) + postfix;
                  }
               }
            }
         }
         
         out.writeBytes(line + LineSeparator);
         i++;
      }
   }
}
