package com.capitalware.ufm.handlers;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.nio.channels.ReadableByteChannel;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.objects.UFMException;

/**
 * This class will handle all of UFM's IO work.
 * @author Roger Lacroix
 * @version 1.1
 * @license Apache 2 License
 */
public class IO
{
   /**
    * The constructor - make sure nothing instantiates this class.
    * It must be accessed in a static way.
    */
   private IO()
   {}

   /**
    * This method will read a file.
    *
    * @param fileName the name of the file.
    * @return byte array of the contents of the file.
    * @throws UFMException
    */
   public static byte[] readFile(String fileName) throws UFMException
   {
      RandomAccessFile in = null;
      FileChannel fileChannel = null;
      FileLock fileLock = null;
      boolean failed = false;
      byte[] byteData = null;

      try
      {
         in = new RandomAccessFile(fileName, "r");
         long len = in.length();
         if (UFM.useFileLocking)
         {
            fileChannel = in.getChannel();
            fileLock = fileChannel.tryLock(0, len, true); /* shared lock for reading */
         }

         if ( ( (fileLock != null) && (UFM.useFileLocking) ) ||
               (!UFM.useFileLocking) )
         {
            byteData = new byte[(int)len];

            in.read(byteData);

            UFM.logger.info("Read completed.");
         }
         else
         {
            UFM.logger.fatal("IO.readFile cannot acquire lock for file: " + fileName);
            UFM.globalErrorText = "IO.readFile cannot acquire lock for file: " + fileName;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if(fileLock != null)
               fileLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (fileChannel != null)
               fileChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (in != null)
               in.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         if (failed == true)
            throw new UFMException("IO.readFile had an error. Please check log for more details.");
      }

      return byteData;
   }

   /**
    * This method will write a file.
    *
    * @param fileName the name of the file.
    * @param data the data to be written.
    * @param append
    * @throws UFMException
    */
   public static void writeFile(String fileName, byte[] data, boolean append) throws UFMException
   {
      RandomAccessFile out = null;
      FileChannel fileChannel = null;
      FileLock fileLock = null;
      boolean failed = false;

      if (!append)
      {
         /* Get rid of the file if it exists. */
         if (new File(fileName).exists())
            deleteFile(fileName);
      }

      try
      {
         out = new RandomAccessFile(fileName, "rw");
         if (UFM.useFileLocking)
         {
            fileChannel = out.getChannel();
            fileLock = fileChannel.tryLock(); /* exclusive lock for writing */
         }

         if ( ( (fileLock != null) && (UFM.useFileLocking) ) ||
              (!UFM.useFileLocking) )
         {
            if (append)
            {
               long fileLen = out.length();
               out.seek(fileLen);   // move to the end of the file.
            }
            out.write(data);
            UFM.logger.info("Successfully completed write to file: "+fileName);
         }
         else
         {
            UFM.logger.fatal("IO.writeFile cannot acquire lock for file: " + fileName);
            UFM.globalErrorText = "IO.writeFile cannot acquire lock for file: " + fileName;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if(fileLock != null)
               fileLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (fileChannel != null)
               fileChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         try
         {
            if (out != null)
               out.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         if (failed == true)
            throw new UFMException("IO.writeFile had an error. Please check log for more details.");
      }

      return;
   }

   /**
    * Append 1 file to the end of another file.
    * @param srcFileName
    * @param dstFileName
    * @throws UFMException
    */
   public static void appendFile(String srcFileName, String dstFileName) throws UFMException
   {
      FileInputStream srcFos = null;
      FileOutputStream dstFos = null;
      FileChannel srcChannel = null;
      FileChannel dstChannel = null;
      FileLock srcLock = null;
      FileLock dstLock = null;
      boolean failed = false;
      long count = 0;
      long position = 0;

      try
      {
         srcFos = new FileInputStream(srcFileName);
         // Create channel on the source
         srcChannel = srcFos.getChannel();
         count = srcChannel.size();
         if (UFM.useFileLocking)
         {
            srcLock = srcChannel.tryLock(0, count, true); /* shared lock for reading */
         }
         dstFos = new FileOutputStream(dstFileName, true);
         // Create channel on the target
         dstChannel = dstFos.getChannel();
         if (UFM.useFileLocking)
         {
            dstLock = dstChannel.tryLock(); /* exclusive lock for writing */
         }
         position = dstChannel.size();


         if ( ( (dstLock != null) && (srcLock != null) && (UFM.useFileLocking) ) ||
              (!UFM.useFileLocking) )
         {
            // Copy file contents from source to destination
            dstChannel.transferFrom(srcChannel, position, count);

            UFM.logger.info("Successfully appended file: "+srcFileName+ " to " + dstFileName);
         }
         else
         {
            String s;
            if (dstLock != null)
               s = "IO.appendFile cannot acquire lock for file: " + dstFileName;
            else
               s = "IO.appendFile cannot acquire lock for file: " + srcFileName;

            UFM.logger.fatal(s);
            UFM.globalErrorText = s;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if(dstLock != null)
               dstLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if(srcLock != null)
               srcLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         // Close the channels
         try
         {
            if (srcChannel != null)
               srcChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstChannel != null)
               dstChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         // Close the files
         try
         {
            if (srcFos != null)
               srcFos.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstFos != null)
               dstFos.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         if (failed == true)
            throw new UFMException("IO.appendFile had an error. Please check log for more details.");
      }
   }

   /**
    * Copy the specified file.
    * @param srcFileName
    * @param dstFileName
    * @throws UFMException
    */
   public static void copyFile(String srcFileName, String dstFileName) throws UFMException
   {
      FileInputStream srcFos = null;
      FileOutputStream dstFos = null;
      FileChannel srcChannel = null;
      FileChannel dstChannel = null;
      FileLock srcLock = null;
      FileLock dstLock = null;
      long len;
      boolean failed = false;

      try
      {
         srcFos = new FileInputStream(srcFileName);
         // Create channel on the source
         srcChannel = srcFos.getChannel();
         len = srcChannel.size();
         if (UFM.useFileLocking)
         {
            srcLock = srcChannel.tryLock(0, len, true); /* shared lock for reading */
         }

         dstFos = new FileOutputStream(dstFileName);
         // Create channel on the destination
         dstChannel = dstFos.getChannel();
         if (UFM.useFileLocking)
         {
            dstLock = dstChannel.tryLock(); /* exclusive lock for writing */
         }

         if ( ( (dstLock != null) && (srcLock != null) && (UFM.useFileLocking) ) ||
              (!UFM.useFileLocking) )
         {
            // Copy file contents from source to destination
            dstChannel.transferFrom(srcChannel, 0, srcChannel.size());

            UFM.logger.info("Successfully copied file: "+srcFileName+ " to " + dstFileName);
         }
         else
         {
            String s;
            if (dstLock != null)
               s = "IO.copyFile cannot acquire lock for file: " + dstFileName;
            else
               s = "IO.copyFile cannot acquire lock for file: " + srcFileName;

            UFM.logger.fatal(s);
            UFM.globalErrorText = s;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if(dstLock != null)
               dstLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if(srcLock != null)
               srcLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         // Close the channels
         try
         {
            if (srcChannel != null)
               srcChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstChannel != null)
               dstChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         // Close the files
         try
         {
            if (srcFos != null)
               srcFos.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstFos != null)
               dstFos.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         if (failed == true)
            throw new UFMException("IO.copyFile had an error. Please check log for more details.");
      }
   }

   /**
    * Copy the specified file from a URL.
    * @param urlFileName
    * @param dstFileName
    * @throws UFMException
    */
   public static void copyFileFromURL(String urlFileName, String dstFileName) throws UFMException
   {
      FileOutputStream dstFos = null;
      FileChannel dstChannel = null;
      FileLock dstLock = null;
      boolean failed = false;
      URL url = null;

      try
      {
         dstFos = new FileOutputStream(dstFileName);
         // Create channel on the target
         dstChannel = dstFos.getChannel();
         if (UFM.useFileLocking)
         {
            dstLock = dstChannel.tryLock(); /* exclusive lock for writing */
         }

         if ( ( (dstLock != null) && (UFM.useFileLocking) ) ||
              (!UFM.useFileLocking) )
         {
            // Copy URL file contents from source to destination
            url = new URL(urlFileName);
            ReadableByteChannel rbc = Channels.newChannel(url.openStream());
            dstChannel.transferFrom(rbc, 0, 1 << 24);

            UFM.logger.info("Successfully copied file: "+urlFileName+ " to " + dstFileName);
         }
         else
         {
            String s = "IO.copyFileFromURL cannot acquire lock for file: " + dstFileName;
            UFM.logger.fatal(s);
            UFM.globalErrorText = s;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if(dstLock != null)
               dstLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstChannel != null)
               dstChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         // Close the files
         try
         {
            if (dstFos != null)
               dstFos.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         if (failed == true)
            throw new UFMException("IO.copyFileFromURL had an error. Please check log for more details.");
      }
   }

   /**
    * Merge 2 or more files to a target file.
    * @param srcFileNames
    * @param dstFileName
    * @throws UFMException
    */
   public static void mergeFiles(String[] srcFileNames, String dstFileName) throws UFMException
   {
      FileInputStream srcFos = null;
      FileOutputStream dstFos = null;
      FileChannel srcChannel = null;
      FileChannel dstChannel = null;
      FileLock srcLock = null;
      FileLock dstLock = null;
      boolean failed = false;
      long count = 0;
      long position = 0;

      try
      {
         dstFos = new FileOutputStream(dstFileName);
         // Create channel on the target
         dstChannel = dstFos.getChannel();
         if (UFM.useFileLocking)
         {
            dstLock = dstChannel.tryLock(); /* exclusive lock for writing */
         }

         if ( ( (dstLock != null) && (UFM.useFileLocking) ) ||
              (!UFM.useFileLocking) )
         {
            for (int i=0; i < srcFileNames.length; i++)
            {
               srcFos = new FileInputStream(srcFileNames[i]);
               // Create channel on the source
               srcChannel = srcFos.getChannel();
               count = srcChannel.size();
               if (UFM.useFileLocking)
               {
                  srcLock = srcChannel.tryLock(0, count, true); /* shared lock for reading */
               }

               if ( ( (srcLock != null) && (UFM.useFileLocking) ) ||
                    (!UFM.useFileLocking) )
               {
                  // Copy file contents from source to destination
                  dstChannel.transferFrom(srcChannel, position, count);

                  position += count;

                  try
                  {
                     if(srcLock != null)
                        srcLock.release();
                  }
                  catch (IOException e)
                  {
                     UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
                     failed = true;
                  }

                  try
                  {
                     // Close the channel
                     if(srcChannel != null)
                        srcChannel.close();
                  }
                  catch (IOException e)
                  {
                     UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
                     failed = true;
                  }

                  try
                  {
                     // Close the file
                     if(srcFos != null)
                        srcFos.close();
                  }
                  catch (IOException e)
                  {
                     UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
                     failed = true;
                  }
               }
               else
               {
                  UFM.logger.fatal("IO.mergeFiles cannot acquire lock for file: " + srcFileNames[i]);
                  UFM.globalErrorText = "IO.mergeFiles cannot acquire lock for file: " + srcFileNames[i];
               }
            }
         }
         else
         {
            UFM.logger.fatal("IO.mergeFiles cannot acquire lock for file: " + dstFileName);
            UFM.globalErrorText = "IO.mergeFiles cannot acquire lock for file: " + dstFileName;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if(dstLock != null)
               dstLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         // Close the channel
         try
         {
            if (dstChannel != null)
               dstChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         // Close the files
         try
         {
            if (dstFos != null)
               dstFos.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         if (failed == true)
            throw new UFMException("IO.mergeFiles had an error. Please check log for more details.");
      }
   }

   /**
    * Delete a file.
    * @param completePathAndFileName
    * @return flag boolean successful or failed
    * @throws UFMException
    */
   public static boolean deleteFile(String completePathAndFileName) throws UFMException
   {
      boolean flag = false;

      File file = new File(completePathAndFileName);
      if (file.delete())
      {
         UFM.logger.info("Successfully deleted file: "+completePathAndFileName);
         flag = true;
      }
      else
      {
         UFM.logger.fatal("Error: Cannot delete file: "+completePathAndFileName);
         UFM.globalErrorText = "Error: Cannot delete file: "+completePathAndFileName;

         flag = false;
         throw new UFMException("IO.deleteFile had an error. Please check log for more details.");
      }

      return flag;
   }

   /**
    * Deletes all files and subdirectories under dir.
    * Returns true if all deletions were successful.
    * If a deletion fails, the method stops attempting to delete and returns false.
    * @param file
    * @return flag boolean successful or failed
    */
   public static boolean deleteDir(File dir)
   {
       if (dir.isDirectory())
       {
           String[] children = dir.list();
           if (children != null)
           {
              for (int i=0; i<children.length; i++)
              {
                  boolean success = deleteDir(new File(dir, children[i]));
                  if (!success)
                  {
                      return false;
                  }
              }
           }
       }

       // The directory is now empty so delete it
       return dir.delete();
   }

   /**
    * Touch a file.
    * @param completePathAndFileName
    * @return flag boolean successful or failed
    * @throws UFMException
    */
   public static boolean touchFile(String completePathAndFileName) throws UFMException
   {
      boolean flag = false;

      File f = new File(completePathAndFileName);

      if (f.exists())
      {
         flag = f.setLastModified((new Date()).getTime());
         UFM.logger.info("Successfully touched file: "+completePathAndFileName);
      }
      else
      {
         try
         {
            flag = f.createNewFile();
            UFM.logger.info("Successfully touched (created) file: "+completePathAndFileName);
         }
         catch(IOException ioe)
         {
            UFM.logger.fatal("Error: Cannot touch (create) file: "+completePathAndFileName + " : "+ioe.getLocalizedMessage());
            flag = false;
            throw new UFMException("IO.touchFile had an error. Please check log for more details.");
         }
      }

      return flag;
   }

   /**
    * Archive a file.
    * @param completePathAndFileName
    * @param targetDir
    * @param targetFilename
    * @throws UFMException
    */
   public static void archiveFile(String completePathAndFileName, String targetDir, String targetFilename) throws UFMException
   {
      String fileName = getFileName(completePathAndFileName);
      String targetCompletePathAndFileName = apendFileToPath(targetDir, (targetFilename!=null?targetFilename:fileName) );

      try
      {
         copyFile(completePathAndFileName, targetCompletePathAndFileName);
      }
      catch (UFMException e)
      {
         throw new UFMException(e.getLocalizedMessage());
      }

      deleteFile(completePathAndFileName);
   }

   /**
    * Split the file by certain number of bytes.
    * @param srcFileName
    * @param toDir
    * @param size
    * @param prefix
    * @param extension
    * @return fileCount number of files created
    * @throws UFMException
    */
   public static int splitByBytes(String srcFileName, String toDir, long size, String prefix, String extension) throws UFMException
   {
      java.text.DecimalFormat df0 = new java.text.DecimalFormat("0000");
      RandomAccessFile  in = null;
      FileChannel       srcChannel = null;
      FileLock          srcLock = null;

      RandomAccessFile  out = null;
      FileChannel       dstChannel = null;
      FileLock          dstLock = null;

      boolean  failed = false;
      String   dstFile = null;
      byte[]   byteData = null;

      long     ss = size;
      long     curCount = 0;
      int      fileCount = 0;

      try
      {
         in = new RandomAccessFile(srcFileName, "r");
         long len = in.length();
         srcChannel = in.getChannel();
         srcLock = srcChannel.tryLock(0, len, true); /* shared lock for reading */

         long fileLen = in.length();

         if (toDir != null)
            dstFile = apendFileToPath(toDir, prefix + "_" + df0.format(fileCount+1));
         else
            dstFile = prefix + "_" + df0.format(fileCount+1);

         if (extension != null)
            dstFile = dstFile + "." + extension;

         out = new RandomAccessFile(dstFile, "rw");
         dstChannel = out.getChannel();
         dstLock = dstChannel.tryLock(); /* exclusive lock for writing */

         fileCount++;

         if ( (dstLock != null) && (srcLock != null) )
         {
            while (true)
            {
               /* check if block is greater than file size. */
               if (ss > fileLen)
                  ss = fileLen;

               /*  Check when it is the last block. */
               if ( (curCount + ss) > fileLen)
                  ss = fileLen - curCount;

               byteData = new byte[(int)ss];

               in.read(byteData);
               out.write(byteData);
               byteData = null; /* tell GC that we are done. */
               curCount += ss;

               if (curCount >= fileLen)
                  break;

               /* Close the current file. */
               try
               {
                  if(dstLock != null)
                     dstLock.release();
               }
               catch (IOException e)
               { UFM.logger.fatal("IOException: dstLock: " + e.getLocalizedMessage()); }

               try
               {
                  if (dstChannel != null)
                     dstChannel.close();
               }
               catch (IOException e)
               { UFM.logger.fatal("IOException: dstChannel: " + e.getLocalizedMessage()); }
               try
               {
                  if (out != null)
                     out.close();
               }
               catch (IOException e)
               { UFM.logger.fatal("IOException: out:" + e.getLocalizedMessage()); }

               /* open the next file. */
               if (toDir != null)
                  dstFile = apendFileToPath(toDir, prefix + "_" + df0.format(fileCount+1));
               else
                  dstFile = prefix + "_" + df0.format(fileCount+1);

               if (extension != null)
                  dstFile = dstFile + "." + extension;

               out = new RandomAccessFile(dstFile, "rw");
               dstChannel = out.getChannel();
               dstLock = dstChannel.tryLock(); /* exclusive lock for writing */

               fileCount++;
            }

            UFM.logger.info("Successfully split file: "+srcFileName+ " into "+ fileCount +" files.");
         }
         else
         {
            String s;
            if (dstLock != null)
               s = "splitByLine cannot acquire lock for file: " + dstFile;
            else
               s = "splitByLine cannot acquire lock for file: " + srcFileName;

            UFM.logger.fatal(s);
            UFM.globalErrorText = s;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if(srcLock != null)
               srcLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (srcChannel != null)
               srcChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (in != null)
               in.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         try
         {
            if(dstLock != null)
               dstLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstChannel != null)
               dstChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         try
         {
            if (out != null)
               out.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         if (failed == true)
            throw new UFMException("IO.unique had an error. Please check log for more details.");
      }

      return fileCount;
   }

   /**
    * Split the file by certain number of lines.
    * @param srcFileName
    * @param toDir
    * @param size
    * @param prefix
    * @param extension
    * @return fileCount number of files created
    * @throws UFMException
    */
   public static int splitByLines(String srcFileName, String toDir, long size, String prefix, String extension) throws UFMException
   {
      String CRLF = System.getProperty("line.separator");
      java.text.DecimalFormat df0 = new java.text.DecimalFormat("0000");

      RandomAccessFile  in = null;
      FileChannel       srcChannel = null;
      FileLock          srcLock = null;

      RandomAccessFile  out = null;
      FileChannel       dstChannel = null;
      FileLock          dstLock = null;

      boolean  failed = false;
      String   dstFile = null;
      long     curCount = 0;
      int      fileCount = 0;
      String   curLine = "";

      try
      {
         in = new RandomAccessFile(srcFileName, "r");
         long len = in.length();
         srcChannel = in.getChannel();
         srcLock = srcChannel.tryLock(0, len, true); /* shared lock for reading */

         if (toDir != null)
            dstFile = apendFileToPath(toDir, prefix + "_" + df0.format(fileCount+1));
         else
            dstFile = prefix + "_" + df0.format(fileCount+1);

         if (extension != null)
            dstFile = dstFile + "." + extension;

         out = new RandomAccessFile(dstFile, "rw");
         dstChannel = out.getChannel();
         dstLock = dstChannel.tryLock(); /* exclusive lock for writing */

         fileCount++;

         if ( (dstLock != null) && (srcLock != null) )
         {
            while ((curLine = in.readLine()) != null)
            {
               if (curCount >= size)
               {
                  /* Close the current file. */
                  try
                  {
                     if(dstLock != null)
                        dstLock.release();
                  }
                  catch (IOException e)
                  { UFM.logger.fatal("IOException: dstLock: " + e.getLocalizedMessage()); }

                  try
                  {
                     if (dstChannel != null)
                        dstChannel.close();
                  }
                  catch (IOException e)
                  { UFM.logger.fatal("IOException: dstChannel: " + e.getLocalizedMessage()); }
                  try
                  {
                     if (out != null)
                        out.close();
                  }
                  catch (IOException e)
                  { UFM.logger.fatal("IOException: out:" + e.getLocalizedMessage()); }

                  /* open the next file. */
                  if (toDir != null)
                     dstFile = apendFileToPath(toDir, prefix + "_" + df0.format(fileCount+1));
                  else
                     dstFile = prefix + "_" + df0.format(fileCount+1);

                  if (extension != null)
                     dstFile = dstFile + "." + extension;

                  out = new RandomAccessFile(dstFile, "rw");
                  dstChannel = out.getChannel();
                  dstLock = dstChannel.tryLock(); /* exclusive lock for writing */

                  fileCount++;
                  curCount = 0; /* reset the count */
               }

               out.writeBytes(curLine);
               out.writeBytes(CRLF);
               curCount++;
            }

            UFM.logger.info("Successfully split file: "+srcFileName+ " into "+ fileCount +" files.");
         }
         else
         {
            String s;
            if (dstLock != null)
               s = "splitByLine cannot acquire lock for file: " + dstFile;
            else
               s = "splitByLine cannot acquire lock for file: " + srcFileName;

            UFM.logger.fatal(s);
            UFM.globalErrorText = s;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if(srcLock != null)
               srcLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (srcChannel != null)
               srcChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (in != null)
               in.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         try
         {
            if(dstLock != null)
               dstLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstChannel != null)
               dstChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         try
         {
            if (out != null)
               out.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         if (failed == true)
            throw new UFMException("IO.unique had an error. Please check log for more details.");
      }

      return fileCount;
   }

   /**
    * Remove adjacent duplicate lines - like 'uniq' in Linux
    * @param srcFileName
    * @param dstFileName
    * @return dupCount number of duplicate lines removed
    * @throws UFMException
    */
   public static int unique(String srcFileName, String dstFileName) throws UFMException
   {
      String CRLF = System.getProperty("line.separator");
      RandomAccessFile in = null;
      FileChannel srcChannel = null;
      FileLock srcLock = null;

      RandomAccessFile out = null;
      FileChannel dstChannel = null;
      FileLock dstLock = null;

      boolean failed = false;
      int dupCount = 0;
      String prevLine = "";
      String curLine = "";

      try
      {
         in = new RandomAccessFile(srcFileName, "r");
         long len = in.length();
         srcChannel = in.getChannel();
         srcLock = srcChannel.tryLock(0, len, true); /* shared lock for reading */

         out = new RandomAccessFile(dstFileName, "rw");
         dstChannel = out.getChannel();
         dstLock = dstChannel.tryLock(); /* exclusive lock for writing */

         if ( (dstLock != null) && (srcLock != null) )
         {
            while ((curLine = in.readLine()) != null)
            {
               /* Only write it out if it is not duplicate of the previous line */
               if (!(prevLine.equals(curLine)))
               {
                  out.writeBytes(curLine);
                  out.writeBytes(CRLF);
                  prevLine = curLine;
               }
               else
                  dupCount++;
            }
         }
         else
         {
            String s;
            if (dstLock != null)
               s = "IO.unique cannot acquire lock for file: " + dstFileName;
            else
               s = "IO.unique cannot acquire lock for file: " + srcFileName;

            UFM.logger.fatal(s);
            UFM.globalErrorText = s;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if(srcLock != null)
               srcLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (srcChannel != null)
               srcChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (in != null)
               in.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         try
         {
            if(dstLock != null)
               dstLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstChannel != null)
               dstChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         try
         {
            if (out != null)
               out.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         if (failed == true)
            throw new UFMException("IO.unique had an error. Please check log for more details.");
      }

      return dupCount;
   }

   /**
    * Remove items in a source file.
    * @param srcFileName
    * @param itemFileName
    * @param dstFileName
    * @throws UFMException
    */
   public static int removeItems(String srcFileName, String itemFileName, String dstFileName) throws UFMException
   {
      String CRLF = System.getProperty("line.separator");
      RandomAccessFile inRAF = null;
      FileChannel srcChannel = null;
      FileLock srcLock = null;

      RandomAccessFile itemRAF = null;
      FileChannel itemChannel = null;
      FileLock itemLock = null;

      RandomAccessFile outRAF = null;
      FileChannel dstChannel = null;
      FileLock dstLock = null;

      boolean failed = false;
      int removeCount = 0;
      String curLine = "";
      Hashtable<String,String> items_HT = new Hashtable<String,String>();

      try
      {
         itemRAF = new RandomAccessFile(itemFileName, "r");
         long len = itemRAF.length();
         itemChannel = itemRAF.getChannel();
         itemLock = itemChannel.tryLock(0, len, true); /* shared lock for reading */

         inRAF = new RandomAccessFile(srcFileName, "r");
         len = inRAF.length();
         srcChannel = inRAF.getChannel();
         srcLock = srcChannel.tryLock(0, len, true); /* shared lock for reading */

         outRAF = new RandomAccessFile(dstFileName, "rw");
         dstChannel = outRAF.getChannel();
         dstLock = dstChannel.tryLock(); /* exclusive lock for writing */

         if ( (itemLock != null) && (dstLock != null) && (srcLock != null) )
         {
            /**
             * Read & load all of the items to be removed from the list.
             */
            while ((curLine = itemRAF.readLine()) != null)
            {
               if (!("".equals(curLine.trim())))
                  items_HT.put(curLine.trim(), "");
            }

            /**
             * Now process the source and remove the items.
             */
            while ((curLine = inRAF.readLine()) != null)
            {
               if (!("".equals(curLine.trim())))
               {
                  if (null == items_HT.get(curLine.trim()))
                  {
                     // not found then output it
                     outRAF.writeBytes(curLine);
                     outRAF.writeBytes(CRLF);
                  }
                  else
                  {
                     removeCount++;
                     UFM.logger.debug("Removing item: "+curLine);
                  }
               }
            }
         }
         else
         {
            String s;
            if (itemLock != null)
               s = "IO.removeItems cannot acquire lock for file: " + itemFileName;
            else if (dstLock != null)
               s = "IO.removeItems cannot acquire lock for file: " + dstFileName;
            else
               s = "IO.removeItems cannot acquire lock for file: " + srcFileName;

            UFM.logger.fatal(s);
            UFM.globalErrorText = s;
         }
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if(srcLock != null)
               srcLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (srcChannel != null)
               srcChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (inRAF != null)
               inRAF.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         try
         {
            if(itemLock != null)
               itemLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (itemChannel != null)
               itemChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (itemRAF != null)
               itemRAF.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         try
         {
            if(dstLock != null)
               dstLock.release();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstChannel != null)
               dstChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
         try
         {
            if (outRAF != null)
               outRAF.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         if (failed == true)
            throw new UFMException("IO.removeItems had an error. Please check log for more details.");
      }

      return removeCount;
   }

   /**
    * Re-Encode a file with a different character set.
    * Take a file that is encoded as <code>srcEncoding</code> and re-encode it
    * as <code>dstEncoding</code>.  The encodings must be one of the strings listed in the Java API docs
    * <a href="http://docs.oracle.com/javase/6/docs/api/java/nio/charset/Charset.html">here</a>.
    *
    * Default Character Sets
    * ASCII, Big5, Big5-HKSCS, CP037, CP1006, CP1025, CP1026, CP1046, CP1097, CP1098, CP1112, CP1122,
    * CP1123, CP1124, CP1140, CP1141, CP1142, CP1143, CP1144, CP1145, CP1146, CP1147, CP1148, CP1149,
    * CP1250, CP1251, CP1252, CP1253, CP1254, CP1255, CP1256, CP1257, CP1258, CP1381, CP1383, CP273,
    * CP277, CP278, CP280, CP284, CP285, CP297, CP33722, CP420, CP424, CP437, CP500, CP737, CP775, CP838,
    * CP850, CP852, CP855, CP856, CP857, CP858, CP860, CP861, CP862, CP863, CP864, CP865, CP866, CP868,
    * CP869, CP870, CP871, CP874, CP875, CP918, CP921, CP922, CP930, CP933, CP935, CP937, CP939, CP942,
    * CP942C, CP943, CP943C, CP948, CP949, CP949C, CP950, CP964, CP970, DBCS-ASCII, DBCS-EBCDIC, EUC,
    * EUC-CN, EUC-JP, EUC-JP-LINUX, EUC-KR, EUC-TW, GBK, HKSCS, ISO2022, ISO2022CN, ISO2022JP, ISO2022KR,
    * ISO8859-1, ISO8859-13, ISO8859-15-FDIS, ISO8859-2, ISO8859-3, ISO8859-4, ISO8859-5, ISO8859-6,
    * ISO8859-7, ISO8859-8, ISO8859-9, JIS0201, JIS0208, JIS0212, JISAutoDetect, Johab, KOI8-R,
    * MacArabic, MacCentralEurope, MacCroatian, MacCyrillic, MacDingbat, MacGreek, MacHebrew, MacIceland,
    * MacRoman, MacRomania, MacSymbol, MacThai, MacTurkish, MacUkraine, MS874, MS932, MS932DB, MS936,
    * MS949, MS950, SJIS, Symbol, TIS620, Unicode, UnicodeBig, UnicodeBigUnmarked, UnicodeLittle,
    * UnicodeLittleUnmarked, UTF-16, UTF-8
    *
    * @param srcFileName
    * @param dstFileName
    * @param srcEncoding
    * @param dstEncoding
    * @throws UFMException
    * @throws Exception
    */
   public static void reencodeFile(String srcFileName, String srcEncoding, String dstFileName, String dstEncoding) throws UFMException
   {
      int chunkSize = 4096;
      boolean failed = false;

      FileInputStream srcFos = null;
      FileOutputStream dstFos = null;
      FileChannel srcChannel = null;
      FileChannel dstChannel = null;
      ByteBuffer srcBuffer = null;
      ByteBuffer dstBuffer = null;
      CharBuffer decodedBuffer = null;

      Charset srcCharset = Charset.forName(srcEncoding);
      Charset dstCharset = Charset.forName(dstEncoding);
      CharsetDecoder decoder = srcCharset.newDecoder();
      CharsetEncoder encoder = dstCharset.newEncoder();

      try
      {
         srcFos = new FileInputStream(srcFileName);
         // Create channel on the source
         srcChannel = srcFos.getChannel();

         dstFos = new FileOutputStream(dstFileName);
         //Create channel on the target
         dstChannel = dstFos.getChannel();

         srcBuffer = ByteBuffer.allocate(chunkSize);
         dstBuffer = ByteBuffer.allocate(chunkSize);
         decodedBuffer = CharBuffer.allocate(chunkSize);

         decoder.reset();
         encoder.reset();
         while( srcChannel.read(srcBuffer) != -1 )
         {
            srcBuffer.flip();
            decoder.decode(srcBuffer, decodedBuffer, false);
            decodedBuffer.flip();
            encoder.encode(decodedBuffer, dstBuffer, false);
            dstBuffer.flip();
            while( dstBuffer.hasRemaining() )
            {
               dstChannel.write(dstBuffer);
            }
            srcBuffer.clear();
            decodedBuffer.clear();
            dstBuffer.clear();
         }

         decoder.decode(srcBuffer, decodedBuffer, true);
         encoder.encode(decodedBuffer, dstBuffer, true);
         decoder.flush(decodedBuffer);
         encoder.flush(dstBuffer);

         UFM.logger.info("Successfully re-encoded file: "+srcFileName+ " to " + dstFileName);
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         failed = true;
      }
      finally
      {
         try
         {
            if (srcChannel != null)
               srcChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstChannel != null)
               dstChannel.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         // Close the files
         try
         {
            if (srcFos != null)
               srcFos.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }

         try
         {
            if (dstFos != null)
               dstFos.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            failed = true;
         }
      }

      if (failed == true)
         throw new UFMException("IO.reencodeFile had an error. Please check log for more details.");
   }

   /**
    * Check an string array of file names for any with a wildcard.  Create a new string array
    * of the file names.
    * @param srcFilenames
    * @return al
    * @throws UFMException
    */
   public static String[] buildStringArrayOfFiles(String[] srcFilenames) throws UFMException
   {
      ArrayList<String> al = new ArrayList<String>();

      for (int i=0; i < srcFilenames.length; i++)
      {
         if (srcFilenames[i].indexOf('*') != -1)
         {
            try
            {
               File[] fileList = IO.getFilteredFileList(srcFilenames[i]);
               for (int j=0; j < fileList.length; j++)
               {
                  al.add(fileList[j].toString());
               }
            }
            catch (IOException e)
            {
               UFM.logger.fatal("IOException: "+e.getLocalizedMessage());
               if (e.getCause() != null)
                  UFM.logger.fatal(e.getCause().toString());
               throw new UFMException("IO.buildStringArrayOfFiles had an error. Please check log for more details.");
            }
         }
         else
         {
            al.add(srcFilenames[i]);
         }
      }

      /**
       * Create string array of the new filenames.
       */
      String[] newFilenames = new String[al.size()];

      for (int i=0; i < al.size(); i++)
         newFilenames[i] = al.get(i);

      return newFilenames;
   }

   /**
    * Retrieve just the filename of a string that may contain both path and filename.
    * @param completePathAndFileName
    * @return fileName
    */
   public static String getFileName(String completePathAndFileName)
   {
      String fileName;

      int x = completePathAndFileName.lastIndexOf(File.separator);
      if (x != -1)
         fileName = completePathAndFileName.substring(x+1);
      else
         fileName = completePathAndFileName;

      return fileName;
   }

   /**
    * Retrieve just the Path of a string that may contain both path and filename.
    * @param completePathAndFileName
    * @return path
    */
   public static String getPath(String completePathAndFileName)
   {
      String path = null;

      if (completePathAndFileName.indexOf(File.separator) != -1)
         path = completePathAndFileName.substring(0,completePathAndFileName.lastIndexOf(File.separator));

      return path;
   }

   /**
    * Build a complete path and filename and check to make sure it has the
    * appropriate file separator.
    *
    * @param path
    * @param filename
    * @return completePathAndFileName
    */
   public static String apendFileToPath(String path, String filename)
   {
      return apendFileToPath(path, filename, File.separator);
   }

   /**
    * Build a complete path and filename and check to make sure it has the
    * appropriate file separator.
    *
    * @param path
    * @param filename
    * @param separator
    * @return completePathAndFileName
    */
   public static String apendFileToPath(String path, String filename, String separator)
   {
      String completePathAndFileName;

      if (path != null)
      {
         if (path.endsWith(separator))
            completePathAndFileName = path + filename;
         else
            completePathAndFileName= path + separator + filename;
      }
      else
         completePathAndFileName = filename;

      return completePathAndFileName;
   }

   /**
    * Retrieve a list of files based on wildcard filter.
    *
    * @param pathAndFileName
    * @return fileList
    */
   public static File[] getFilteredFileList(String pathAndFileName) throws IOException
   {
      String pathOnly;
      String filter;

      if (pathAndFileName.lastIndexOf(File.separator) != -1)
      {
         pathOnly = pathAndFileName.substring(0,pathAndFileName.lastIndexOf(File.separator));
         filter   = pathAndFileName.substring(pathAndFileName.lastIndexOf(File.separator) + 1);
      }
      else
      {
         pathOnly   = System.getProperty("user.dir");
         filter = pathAndFileName;
      }

      File dir = new File(pathOnly);

      if (!(dir.exists()))
            throw new IOException("Invalid directory : " + pathOnly);

      return dir.listFiles(new FilterFiles(filter));
   }

   /**
    * Copies an input stream to an output stream.
    *
    * @param in
    *           The InputStream.
    * @param out
    *           The OutputStream.
    */
   public static void streamCopier(InputStream in, OutputStream out) throws IOException
   {
      synchronized (in)
      {
         synchronized (out)
         {
            byte[] buffer = new byte[4096];
            while (true)
            {
               int bytesRead = in.read(buffer);
               if (bytesRead == -1)
                  break;
               out.write(buffer, 0, bytesRead);
            }
         }
      }
   }
}
