package com.capitalware.ufm.mq;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;

import org.jdom.JDOMException;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.handlers.WorkerThreadForPA;
import com.capitalware.ufm.action.objects.UFMObject;
import com.capitalware.ufm.action.objects.MQReceiveObject;
import com.capitalware.ufm.handlers.AES;
import com.capitalware.ufm.handlers.IO;
import com.capitalware.ufm.handlers.InternalMessageQueue;
import com.capitalware.ufm.handlers.RemoteValuesInMsg;
import com.capitalware.ufm.handlers.RunJob;
import com.capitalware.ufm.handlers.JZip;
import com.capitalware.ufm.handlers.StringUtilities;
import com.capitalware.ufm.internal.messages.UpdateStatus;
import com.capitalware.ufm.mq.objects.MQObject;
import com.capitalware.ufm.mq.objects.QueueObject;
import com.capitalware.ufm.objects.CWConstants;
import com.capitalware.ufm.objects.UFMException;

import com.ibm.mq.*;
import com.ibm.mq.constants.CMQC;

/**
 * This class will handle all MQ interactions for getting message(s) and writing 
 * it to a file.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class MQGetMsg extends MyMQ
{
   private        boolean        loopAgain = true;
//   private        boolean        hasStop = false;
   private        int            msgCounter = 0;
   private        MQQueue        inQ = null;
   private        MQQueue        boQ = null;

   /**
    * The constructor
    */
   public MQGetMsg(MQObject qm)
   {
   	super(qm);
   }

   /**
    * Exit loop processing
    */
   public void stopIt()
   {
      loopAgain = false;
   }
   
   /**
    * Retrieve a message from a queue and write it to a file.
    * @param ro
    * @throws UFMException 
    */
   public int retrieveMsg(MQReceiveObject ro, InternalMessageQueue iq) throws UFMException
   {
      int exitCode = 0;
      int xx = 1;
      MQGetMessageOptions gmo = new MQGetMessageOptions();
      gmo.matchOptions = CMQC.MQMO_NONE;
      gmo.options = CMQC.MQGMO_FAIL_IF_QUIESCING + CMQC.MQGMO_SYNCPOINT + CMQC.MQGMO_LOGICAL_ORDER;

      if (ro.getWithConvert)
         gmo.options += CMQC.MQGMO_CONVERT;

      if (ro.processType == CWConstants.DAEMON)
      {
         gmo.options += CMQC.MQGMO_WAIT;
         gmo.waitInterval = 2000; // wait up to 2 seconds
      }

      inQ = openInQ(mqo.qObject);

      if (ro.backOutQName != null)
         boQ = openOutQ(ro.backOutQName, false);

      if ( (inQ == null) || 
           ( (ro.backOutQName != null) && (boQ == null) ) )
         loopAgain = false;

      MQMessage receiveMsg = new MQMessage();
      receiveMsg.messageId     = CMQC.MQMI_NONE;
      receiveMsg.correlationId = CMQC.MQCI_NONE;
      receiveMsg.groupId       = CMQC.MQGI_NONE;

      try
      {
         while (loopAgain)
         {
            try
            {
               inQ.get(receiveMsg, gmo);
               ro.msgCount++;
   
               if (receiveMsg.feedback == CMQC.MQFB_QUIT)
               {
                  loopAgain = false;
                  break;
               }
               
               if ((receiveMsg.messageFlags & CMQC.MQMF_MSG_IN_GROUP) == CMQC.MQMF_MSG_IN_GROUP)
               {
                  handleGroupOfMsgs(ro, receiveMsg, inQ, boQ, iq);
               }
               else
               {
                  handleOutputtingSingleFile(ro, receiveMsg, boQ, iq);
               }
               
               receiveMsg.clearMessage();
               receiveMsg.messageId     = CMQC.MQMI_NONE;
               receiveMsg.correlationId = CMQC.MQCI_NONE;
               receiveMsg.groupId       = CMQC.MQGI_NONE;
               xx = 1; // reset the counter
            }
            catch (JDOMException je)
            {
               UFM.logger.fatal("JDOMException: " + je.getLocalizedMessage());
               if (je.getCause() != null)
                  UFM.logger.fatal(je.getCause().toString());
               UFM.globalErrorText = je.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (MQException e)
            {
               if ( (ro.processType == CWConstants.TRIGGERED)    &&
                    (e.completionCode == CMQC.MQCC_FAILED)       &&
                    (e.reasonCode == CMQC.MQRC_NO_MSG_AVAILABLE) )
               {
                  UFM.logger.info("Triggering loop : no more messages.");
                  loopAgain = false;
               }
               else if ( (ro.processType == CWConstants.ONE_PASS)  &&
                     (e.completionCode == CMQC.MQCC_FAILED)        &&
                     (e.reasonCode == CMQC.MQRC_NO_MSG_AVAILABLE)  )
                {
                   UFM.logger.info("One Pass loop : no more messages.");
                   loopAgain = false;
                }
               else if ( (ro.processType == CWConstants.RUN_UNTIL_EMPTY) &&
                         (e.completionCode == CMQC.MQCC_FAILED)          &&
                         (e.reasonCode == CMQC.MQRC_NO_MSG_AVAILABLE)    )
                {
                   UFM.logger.info("RunUntilEmpty loop : no more messages.");
                   loopAgain = false;
                }
               else if ( (ro.processType == CWConstants.DAEMON)       &&
                         (e.completionCode == CMQC.MQCC_FAILED)       &&
                         (e.reasonCode == CMQC.MQRC_NO_MSG_AVAILABLE) )
               {
                  if (xx >= 30)
                  {
                     // output this message once every minute
                     UFM.logger.info("Daemon loop : no message found.");
                     xx = 1; // reset the counter
                  }
                  else
                     xx++;
               }
               else if ( (e.completionCode == CMQC.MQCC_FAILED)                 &&
                         ( (e.reasonCode == CMQC.MQRC_CONNECTION_QUIESCING) ||
                           (e.reasonCode == CMQC.MQRC_Q_MGR_QUIESCING)      ||
                           (e.reasonCode == CMQC.MQRC_Q_MGR_STOPPING)       ||
                           (e.reasonCode == CMQC.MQRC_CONNECTION_BROKEN)    )    )
               {
                  UFM.logger.fatal("MQException: " + e.getLocalizedMessage() + " : QName=" + mqo.qObject);
                  UFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);
                  
                  if (!reconnect(ro))  // attempt to reconnect to the queue manager
                  {
                     loopAgain = false;
                     exitCode = 99;
                  }
               }
               else
               {
                  UFM.logger.fatal("MQException: " + e.getLocalizedMessage() + " : QName=" + mqo.qObject);
                  UFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);

                  if (!reconnect(ro))  // attempt to reconnect to the queue manager
                  {
                     UFM.globalErrorText = e.getLocalizedMessage();
                     loopAgain = handleProblem(ro, receiveMsg, boQ);
                     if (!loopAgain)
                        exitCode = 99;
                  }
               }
            }
            catch (UFMException e)
            {
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (IOException e)
            {
               UFM.logger.fatal("IOException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (InvalidKeyException e)
            {
               UFM.logger.fatal("InvalidKeyException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (NoSuchAlgorithmException e)
            {
               UFM.logger.fatal("NoSuchAlgorithmException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (NoSuchPaddingException e)
            {
               UFM.logger.fatal("NoSuchPaddingException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (InvalidAlgorithmParameterException e)
            {
               UFM.logger.fatal("InvalidAlgorithmParameterException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (IllegalBlockSizeException e)
            {
               UFM.logger.fatal("IllegalBlockSizeException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (BadPaddingException e)
            {
               UFM.logger.fatal("BadPaddingException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }

            if (ro.processType == CWConstants.ONE_PASS)
               loopAgain = false;
         }
      }
      finally
      {
         closeQ(inQ);
         closeQ(boQ);
      }

      UFM.logger.info("Receive Action processed "+msgCounter+" messages.");

      return exitCode;
   }

   /**
    * Retrieve all messages in a group and write them to a file.
    * @param ro
    * @param receiveMsg
    * @param msgCounter
    * @param inQ
    * @param boQ
    * @throws IOException 
    * @throws JDOMException 
    * @throws UFMException 
    * 
    */
   private void handleGroupOfMsgs(MQReceiveObject ro, MQMessage receiveMsg, MQQueue inQ, MQQueue boQ, InternalMessageQueue iq) throws JDOMException, IOException, UFMException
   {
      SimpleDateFormat dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      boolean loopMsgInGroup = true;
      boolean processIt = true;
      int msgNum = 1;
      long fileSize = 0;
      byte[] b;
      RandomAccessFile out = null;
      FileChannel fileChannel = null;
      FileLock fileLock = null;
      
      QueueObject qo = new QueueObject(receiveMsg.replyToQueueManagerName, receiveMsg.replyToQueueName);

      RemoteValuesInMsg rv = new RemoteValuesInMsg();
      rv.processValues(ro, receiveMsg, msgCounter, ro.paClass.actionProperties);
      
      MQGetMessageOptions gmo = new MQGetMessageOptions();
      gmo.matchOptions = CMQC.MQMO_MATCH_GROUP_ID;
      gmo.waitInterval = 5000;  // 5 Seconds
      gmo.options = CMQC.MQGMO_FAIL_IF_QUIESCING + CMQC.MQGMO_SYNCPOINT + CMQC.MQGMO_WAIT + CMQC.MQGMO_LOGICAL_ORDER;

      if (ro.getWithConvert)
         gmo.options += CMQC.MQGMO_CONVERT;

      /* Create the directory if it does not exist? */
      if (ro.createDir)
      {
         if (!(new File(rv.getDirectory())).exists())
         {
            if ((new File(rv.getDirectory())).mkdirs()) 
               UFM.logger.info("Successfully created directory: " + rv.getDirectory());
            else
               throw new IOException("handleGroupOfMsgs(): Cannot create directory: " + rv.getDirectory());
         }
      }
      
      try
      {
         out = new RandomAccessFile(rv.getToPathAndFilename(), "rw");
         fileChannel = out.getChannel();
         fileLock = fileChannel.tryLock(); /* exclusive lock for writing */
         
         if (fileLock == null)
         {
            UFM.logger.fatal("handleGroupOfMsgs cannot acquire lock for file: " + rv.getToPathAndFilename());
            UFM.globalErrorText = "handleGroupOfMsgs cannot acquire lock for file: " + rv.getToPathAndFilename();
            throw new UFMException("handleGroupOfMsgs cannot acquire lock for file: " + rv.getToPathAndFilename());
         }
         
         if ( ((receiveMsg.feedback & CWConstants.UFM_ENCRYPT_MSG) == CWConstants.UFM_ENCRYPT_MSG) &&
               (ro.crypto == null) )
         {
            UFM.logger.error("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
            if (handleProblem(ro, receiveMsg, boQ))
               processIt = false;
            else
               throw new UFMException("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
         }
         
         while (loopMsgInGroup)
         {
            try
            {
               if (processIt)
               {
                  if ( (msgNum == 1) &&
                       (receiveMsg.feedback & CWConstants.UFM_XML_EMBEDDED_MSG) == CWConstants.UFM_XML_EMBEDDED_MSG)
                     b = new byte[receiveMsg.getMessageLength() - rv.getXmlDataLength()];
                  else
                     b = new byte[receiveMsg.getMessageLength()];
                  
                  receiveMsg.readFully(b);
                  msgCounter++;
                  
                  if ((receiveMsg.feedback & CWConstants.UFM_ZIP_ENCRYPT_MSG) == CWConstants.UFM_ZIP_ENCRYPT_MSG)
                  {
                     AES aes = new AES(ro.crypto.passPhrase, ro.crypto.keySize);
                     b = JZip.decompressData(aes.decrypt(b));
                  }
                  else if ((receiveMsg.feedback & CWConstants.UFM_ENCRYPT_MSG) == CWConstants.UFM_ENCRYPT_MSG)
                  {
                     AES aes = new AES(ro.crypto.passPhrase, ro.crypto.keySize);
                     b = aes.decrypt(b);
                  }
                  else if ((receiveMsg.feedback & CWConstants.UFM_ZIP_MSG) == CWConstants.UFM_ZIP_MSG)
                     b = JZip.decompressData(b);
                     
                  out.write(b);
                  fileSize += b.length;
      
                  _qMgr.commit();
      
                  if ((receiveMsg.messageFlags & CMQC.MQMF_LAST_MSG_IN_GROUP) == CMQC.MQMF_LAST_MSG_IN_GROUP)
                     UFM.logger.info("Retrieved last message # "+msgNum+" and wrote it to '" + rv.getToPathAndFilename() + "'" );
                  else
                     UFM.logger.info("Retrieved message # "+msgNum+" and wrote it to '" + rv.getToPathAndFilename() + "'" );
                  
                  if (iq != null)
                     iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+ro.actionName+", "+ro.labelName+", MQGet, Event=MQ Message "+msgNum+" written to "+rv.getToPathAndFilename()+", \n"));
               }
               
               /**
                * Setup and get the next message in group.
                */
               processIt = true;
               receiveMsg.clearMessage();
               receiveMsg.messageId     = CMQC.MQMI_NONE;
               receiveMsg.correlationId = CMQC.MQCI_NONE;
               
               inQ.get(receiveMsg, gmo);
               msgNum++;
               ro.msgCount++;

               if ( ((receiveMsg.feedback & CWConstants.UFM_ENCRYPT_MSG) == CWConstants.UFM_ENCRYPT_MSG) &&
                     (ro.crypto == null) )
               {
                  UFM.logger.error("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
                  if (handleProblem(ro, receiveMsg, boQ))
                     processIt = false;
                  else
                     throw new UFMException("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
               }
            }
            catch (MQException e)
            {
               if (e.reasonCode == CMQC.MQRC_NO_MSG_AVAILABLE)
               {
                  loopMsgInGroup = false;
               }
               else
               {
                  UFM.logger.fatal("MQException: " + e.getMessage() + " : QName=" + mqo.qObject);
                  UFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);

                  UFM.globalErrorText = e.getLocalizedMessage();
                  loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
               }
            }
            catch (IOException e)
            {
               UFM.logger.fatal("IOException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();

               loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
            }
            catch (InvalidKeyException e)
            {
               UFM.logger.fatal("InvalidKeyException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();

               loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
            }
            catch (NoSuchAlgorithmException e)
            {
               UFM.logger.fatal("NoSuchAlgorithmException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();

               loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
            }
            catch (NoSuchPaddingException e)
            {
               UFM.logger.fatal("NoSuchPaddingException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();

               loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
            }
            catch (InvalidAlgorithmParameterException e)
            {
               UFM.logger.fatal("InvalidAlgorithmParameterException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();

               loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
            }
            catch (IllegalBlockSizeException e)
            {
               UFM.logger.fatal("IllegalBlockSizeException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();

               loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
            }
            catch (BadPaddingException e)
            {
               UFM.logger.fatal("BadPaddingException: " + e.getMessage());
               UFM.globalErrorText = e.getLocalizedMessage();

               loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
            }
         }
         
         ro.fileCount++;

         UFM.logger.info("Retrieved group of messages and they were written to '" + rv.getToPathAndFilename() + "' and filesize is "+fileSize);
      }
      finally
      {
         try
         {
            if(fileLock != null) 
               fileLock.release();
         }
         catch (IOException e)
         {           
            UFM.logger.fatal("IOException: " + e.getMessage());
            UFM.globalErrorText = e.getLocalizedMessage();
         }

         try
         {
            if (fileChannel != null)
               fileChannel.close();
         }
         catch (IOException e)
         {           
            UFM.logger.fatal("IOException: " + e.getMessage());
            UFM.globalErrorText = e.getLocalizedMessage();
         }

         try
         {
            if (out != null)
               out.close();
         }
         catch (IOException e)
         {           
            UFM.logger.fatal("IOException: " + e.getMessage());
            UFM.globalErrorText = e.getLocalizedMessage();
         }
      }
      
      handlePostCommon(ro, rv, iq, qo);
   }

   /**
    * Output a single message to a file
    * @param ro
    * @param receiveMsg
    * @param rv
    * @param msgCounter
    * @throws IOException 
    * @throws MQException 
    * @throws JDOMException 
    * @throws InvalidAlgorithmParameterException 
    * @throws NoSuchPaddingException 
    * @throws NoSuchAlgorithmException 
    * @throws InvalidKeyException 
    * @throws BadPaddingException 
    * @throws IllegalBlockSizeException 
    * @throws UFMException 
    */
   private void handleOutputtingSingleFile(MQReceiveObject ro, MQMessage receiveMsg, MQQueue boQ, InternalMessageQueue iq) throws IOException, MQException, JDOMException, InvalidKeyException, NoSuchAlgorithmException, NoSuchPaddingException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException, UFMException
   {
      SimpleDateFormat dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      boolean processIt = true;
      byte[] b;

      UFM.logger.debug("CorrelID_HEX="+StringUtilities.byteArrayToHexString(receiveMsg.correlationId));
      
      QueueObject qo = new QueueObject(receiveMsg.replyToQueueManagerName, receiveMsg.replyToQueueName);

      RemoteValuesInMsg rv = new RemoteValuesInMsg();
      rv.processValues(ro, receiveMsg, msgCounter, ro.paClass.actionProperties);

      if ( ((receiveMsg.feedback & CWConstants.UFM_ENCRYPT_MSG) == CWConstants.UFM_ENCRYPT_MSG) &&
           (ro.crypto == null) )
      {
         UFM.logger.error("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
         if (handleProblem(ro, receiveMsg, boQ))
            processIt = false;
         else
            throw new UFMException("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
      }
      
      if (processIt)
      {
         /* Create the directory if it does not exist? */
         if (ro.createDir)
         {
            if (!(new File(rv.getDirectory())).exists())
            {
               if ((new File(rv.getDirectory())).mkdirs()) 
                  UFM.logger.info("Successfully created directory: " + rv.getDirectory());
               else
                  throw new IOException("handleOutputtingSingleFile(): Cannot create directory: " + rv.getDirectory());
            }
         }
         
         if ((receiveMsg.feedback & CWConstants.UFM_XML_EMBEDDED_MSG) == CWConstants.UFM_XML_EMBEDDED_MSG)
            b = new byte[receiveMsg.getMessageLength() - rv.getXmlDataLength()];
         else
            b = new byte[receiveMsg.getMessageLength()];
         
         receiveMsg.readFully(b);
         
         if ((receiveMsg.feedback & CWConstants.UFM_ZIP_ENCRYPT_MSG) == CWConstants.UFM_ZIP_ENCRYPT_MSG)
         {
            AES aes = new AES(ro.crypto.passPhrase, ro.crypto.keySize);
            try
            {
               IO.writeFile(rv.getToPathAndFilename(), JZip.decompressData(aes.decrypt(b)), ro.append);
            }
            catch (UFMException e)
            {
               if (ro.onErrorFail)
                  throw e;
               else
                  UFM.handleOnError(ro.paClass.actionProperties, iq);
            }
         }
         else if ((receiveMsg.feedback & CWConstants.UFM_ENCRYPT_MSG) == CWConstants.UFM_ENCRYPT_MSG)
         {
            AES aes = new AES(ro.crypto.passPhrase, ro.crypto.keySize);
            try
            {
               IO.writeFile(rv.getToPathAndFilename(), aes.decrypt(b), ro.append);
            }
            catch (UFMException e)
            {
               if (ro.onErrorFail)
                  throw e;
               else
                  UFM.handleOnError(ro.paClass.actionProperties, iq);
            }
         }
         else if ((receiveMsg.feedback & CWConstants.UFM_ZIP_MSG) == CWConstants.UFM_ZIP_MSG)
         {
            try
            {
               IO.writeFile(rv.getToPathAndFilename(), JZip.decompressData(b), ro.append);
            }
            catch (UFMException e)
            {
               if (ro.onErrorFail)
                  throw e;
               else
                  UFM.handleOnError(ro.paClass.actionProperties, iq);
            }
         }
         else
         {
            try
            {
               IO.writeFile(rv.getToPathAndFilename(), b, ro.append);
            }
            catch (UFMException e)
            {
               if (ro.onErrorFail)
                  throw e;
               else
                  UFM.handleOnError(ro.paClass.actionProperties, iq);
            }
         }

         if (iq != null)
            iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+ro.actionName+", "+ro.labelName+", MQGet, Event=MQ Message written to "+rv.getToPathAndFilename()+"\n"));
         
         UFM.logger.info("Retrieved message written to '" + rv.getToPathAndFilename() + "'" );
         msgCounter++;
         ro.fileCount++;
   
         _qMgr.commit();
         
         handlePostCommon(ro, rv, iq, qo);
      }
   }

   /**
    * Handle the common items after the message/file has been received.
    * @param ro
    * @param rv
    * @param iq
    * @throws IOException 
    * @throws JDOMException 
    */
  private void handlePostCommon(MQReceiveObject ro, RemoteValuesInMsg rv, InternalMessageQueue iq, QueueObject qo) throws JDOMException, IOException
  {
     /* Clone the user properties */
     Hashtable<String,String> propClone = (Hashtable<String,String>) ro.paClass.actionProperties.clone();
     /* Put the Watch Found file in the cloned properties */
     propClone.put(CWConstants.MQ_RECEIVE_FILE, rv.getToPathAndFilename());
     propClone.put(CWConstants.MQ_REPLY_TO_QMGR_NAME, qo.qMgrName);
     propClone.put(CWConstants.MQ_REPLY_TO_QUEUE_NAME, qo.qName);

     if (rv.isRunJob())
     {
        RunJob rj = new RunJob(iq);
        rj.launch(rv.getJobXmlFile(), propClone, ro.actionName, ro.labelName);
     }

     if (ro.actionAL != null)
     {
        /* Clone the array of objects */
        ArrayList<UFMObject> newActionAL = new ArrayList<UFMObject>();
        for (int k=0; k < ro.actionAL.size(); k++)
           newActionAL.add((UFMObject)ro.actionAL.get(k).clone());
   
        /* To expedite the work - launch each WatchItem's Actions in their own thread. */
        WorkerThreadForPA wt = new WorkerThreadForPA(newActionAL,
                                                     propClone,
                                                     iq);
        wt.start();
     }
  }

   /**
    *
    * @param ro
    * @param receiveMsg
    * @param boQ
    * @return
    */
   private boolean handleProblem(MQReceiveObject ro, MQMessage receiveMsg, MQQueue boQ)
   {
      boolean flag = true;

      /*
       * Do we have a backout queue?
       * If so, attempt to move the message to it.
       */
      if (boQ != null)
      {
         UFM.logger.info("Attempting to move the message to the backout queue: " + ro.backOutQName);
         MQPutMessageOptions pmo = new MQPutMessageOptions();
         pmo.options = CMQC.MQPMO_FAIL_IF_QUIESCING;
         try
         {
            MQMessage sendmsg = receiveMsg;
            // put the message on the queue.
            boQ.put(sendmsg, pmo);
            _qMgr.commit();
            UFM.logger.info("Successfully moved the message to the backout queue: " + ro.backOutQName);
         }
         catch (MQException e1)
         {
            UFM.logger.fatal("MQException: " + e1.getMessage() + " : QName=" + mqo.qObject);
            UFM.logger.fatal("MQCC=" + e1.completionCode + " : MQRC="+e1.reasonCode);
            UFM.globalErrorText = e1.getLocalizedMessage();

            backout();
            flag = false; 
         }
      }
      else
      {
         backout();
         flag = false; 
      }

      return flag;
   }

   /**
    * Attempt to reconnect to the queue manager.
    */
   private boolean reconnect(MQReceiveObject ro)
   {
      boolean more = true;

      UFM.logger.debug("MQGetMsg(): reconnect()");
      
      // Done. Close the queues.
      if (inQ != null)
      {
         closeQ(inQ);
         inQ = null;
      }

      if (boQ != null)
      {
         closeQ(boQ);
         boQ = null;
      }

      // Disconnect from the queue manager.
      disconnect();

      while (more  && loopAgain)
      {
         // sleep for 60 seconds - 1 second at a time just in case the service 
         // needs to be shutdown while we are sleeping 
         for (int i=0; i < 60; i++)
         {
            try
            {
               Thread.sleep(1000); // time in milliseconds
            }
            catch (InterruptedException ie) {}
            
            if (!loopAgain)  // is this thread suppose to stop?
            {
               more = false;
               break;
            }
         }

         try
         {
            connect();

            inQ = openInQ(mqo.qObject);

            if (ro.backOutQName != null)
               boQ = openOutQ(ro.backOutQName, false);

            if ( (inQ == null) ||
                 ( (ro.backOutQName != null) && (boQ == null) ) )
               more = true;
            else
               more = false;
         }
         catch (MQException e)
         {
            UFM.logger.fatal("MQException: " + e.getLocalizedMessage());
            UFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);
         }
         catch (Exception e)
         {
            UFM.logger.fatal("Exception: " + e.getMessage());
         }
         finally
         {
            // if more is true then the open may have failed - hence disconnect and try again.
            if (more)
            {
               if (inQ != null)
               {
                  closeQ(inQ);
                  inQ = null;
               }

               if (boQ != null)
               {
                  closeQ(boQ);
                  boQ = null;
               }

               disconnect();
            }
         }
      }

      if (more)
         return false;  // failed to reconnect
      else
         return true;   // reconnection successful
   }
}
