package com.capitalware.ufm.watch;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.ArrayList;
import java.util.Hashtable;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.handlers.WorkerThreadForPA;
import com.capitalware.ufm.action.objects.UFMObject;
import com.capitalware.ufm.handlers.InternalMessageQueue;
import com.capitalware.ufm.objects.CWConstants;

/**
 * TimerListener implements an ActionListener interface and listens to the
 * main timer. Each time the timer fires an action Event, all defined watches
 * are checked and the desired file is put to the queue.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class TimerListener implements ActionListener
{
   private Watch[]                  watch = null;
   private Hashtable<String,String> properties = null;
   private boolean                  working = false;
   private InternalMessageQueue             iq = null;

   /**
    * @param watch
    *           Array of watches containing the watch information as Watch
    *           objects
    * @param mailinfo
    *           mail related information
    */
   public TimerListener(Watch[] watch, Hashtable<String,String> properties, InternalMessageQueue iq)
   {
      super();
      this.watch = watch;
      this.properties = properties;
      this.iq = iq;
   }

   /**
    * <P>
    * This method is called from the Timer instance (UFM) every time
    * the polling interval has been finished.
    * </P>
    *
    * <P>
    * Then every time the actionEvent is fired, all defined watches are checked
    * and if a match is found, push the file.
    * </P>
    */
   public void actionPerformed(ActionEvent ae)
   {
      String watchFile = null;
      Hashtable<String,String> propClone = null;
      UFM.logger.debug("TimerListener: actionPerformed");
      
      working = false;
      for (int i = 0; i < watch.length; i++)
      {
         ArrayList<String> al = watch[i].check();
         int al_size = al.size(); 
         if (al_size > 0)
         {
            UFM.logger.info("There is/are "+al_size+" item(s).");
            working = true;

            for (int j=0; j < al_size; j++)
            {
               watchFile = al.get(j);
               UFM.logger.info("Watched file found: " + watchFile);

               if (watch[i].actionAL != null)
               {
                  /* Clone the user properties */ 
                  propClone = (Hashtable<String,String>) properties.clone();
                  /* Put the Watch Found file in the cloned properties */
                  propClone.put(CWConstants.WATCH_FOUND_FILE, watchFile);

                  /* Clone the array of objects */
                  ArrayList<UFMObject> newActionAL = new ArrayList<UFMObject>();
                  for (int k=0; k < watch[i].actionAL.size(); k++)
                     newActionAL.add((UFMObject)watch[i].actionAL.get(k).clone());
                  
                  /* To expedite the work - launch each WatchItem's Actions in their own thread. */ 
                  WorkerThreadForPA wt = new WorkerThreadForPA(newActionAL, 
                                                               propClone, 
                                                               watch[i],
                                                               iq); 
                  wt.start();

                  /* don't sleep on the last one */
                  if (j != (al_size - 1))
                  {
                     /* Put a slight delay in */
                     try {Thread.sleep(85);} catch (InterruptedException e){}
                  }
               }
            }
            
            UFM.logger.debug("TimerListener: actionPerformed: after try/catch");
         }
      }
      working = false;
      UFM.logger.debug("TimerListener: actionPerformed: end");
   }

   /**
    * Are in the process of sending files?
    * @return true / false
    */
   public boolean isWorking()
   {
      return working;
   }
}
